import logging
import json
import time
from TDhelper.network.http.REST_HTTP import GET
from TDhelper.generic.dynamic.base.Meta import dynamic_creator
from TDhelper.network.rpc.Core.struct import RPC_SERVICE
from TDhelper.Decorators.log import logging, logging_setup
from TDhelper.Decorators.performance import performance

performance.on_off(True)


class client():
    __TOKEN_KEY__ = "api-token"
    __TOKEN__ = ""
    __LOGGER__ = logging

    @performance.performance_testing
    def __init__(self, uri, path: list = [], token: str = "", sniffer: list = ["sniffer"], services: list = [], token_key=None, logger: dict = None, try_count=5, try_sleep=3):
        if logger:
            self.__LOGGER__.basicConfig(level=logging.INFO)
            self.__LOGGER__.config.dictConfig(logger)
        self.__TOKEN_KEY__ = token_key if token_key else self.__TOKEN_KEY__
        self.__TOKEN__ = token
        header = {
            self.__TOKEN_KEY__: self.__TOKEN__
        }
        uri = uri.rstrip("/")
        rpc_uri = uri+"/"+"/".join(path)
        if services:
            rpc_uri += "?key="+",".join(services)
        rpc_server_state = False
        for v in range(0, try_count):
            body = self.__get_conf__(uri+"/"+"/".join(sniffer), header)
            if body:
                if body['state'] == 200:
                    rpc_server_state = True
                    break
            time.sleep(try_sleep)
        if not rpc_server_state:
            raise Exception("can not connect rpc server(%s)." %
                            (uri+"/"+"/".join(sniffer)))
        self.__context__ = self.__get_conf__(rpc_uri, header)
        if self.__context__:
            if self.__context__["msg"]:
                self.__dict__["__srv_struct__"] = {}
                for k, v in self.__context__["msg"].items():
                    self.__dict__["__srv_struct__"][k] = type(
                        k, (RPC_SERVICE,), {}).create_by_cnf(v, self.__TOKEN_KEY__)
                    methods = [kv for kv in v['methods'].keys()]
                    self.__dict__[k] = type(k, (dynamic_creator,), {
                                            "__dynamic_methods__": methods, "__hook_method__": self.__call__})()
            else:
                self.__LOGGER__.error("not found rpc method.")
                raise Exception("not found rpc method.")
        else:
            raise Exception("not found rpc instance config.")
        super(client, self).__init__()

    @performance.performance_testing
    def __call__(self, service_name, fun_name, *args, **kwargs):
        remote_func_name = ("".join([service_name, '.', fun_name]))
        if service_name in self.__srv_struct__:
            if hasattr(self.__srv_struct__[service_name], 'methods'):
                if fun_name in self.__srv_struct__[service_name].methods:
                    try:
                        start = time.clock_gettime(0)
                        result = self.__srv_struct__[service_name].__remote_call__(
                            fun_name, *args, **kwargs)
                        end = time.clock_gettime(0)
                        self.__LOGGER__.info("call %s, time consuming %f(s)" % (
                            remote_func_name, (end-start)))
                        return result
                    except Exception as e:
                        self.__LOGGER__.error(e)
                else:
                    self.__LOGGER__.error(
                        "in service(%s) not found method(%s)", (service_name, fun_name))
            else:
                self.__LOGGER__.error("service obj not found key(methods).")
        else:
            self.__LOGGER__.error('not found service(%s).' % service_name)
        return {"state": -1, "msg": {}}

    def __get_conf__(self, uri, header={}):
        state, body = GET(uri, http_headers=header,
                          time_out=5, charset="utf-8")
        if state == 200:
            body = json.loads(body)
            if body["state"] == 200:
                return body
            else:
                return {}
        return {}



from copy import deepcopy
import random
from TDhelper.network.http.REST_HTTP import GET


class HostManager:
    def __init__(self) -> None:
        self.__host_cnf_backup__ = {}
        self.__host_cnf__ = {}
        self.__host__ = []
        self.__host_len__ = 0
        # self.__health_sniffer__ = ""

    def __backup_host__(self):
        self.__host_cnf_backup__ = deepcopy(self.__host_cnf__)

    def get_per_http_host(self):
        try:
            return self.__generateHost__(self.getHost())
        except Exception as e:
            raise e

    def getHost(self):
        """getHost
            get an access host cnf.
        Parameters:
            self - <class: HostManager>, hostManager instance.
        Returns:
            success - <class, json>, host cnf json.
            fail - None.
        Raises:
            None
        """
        result = None
        retry = 3
        while (True and retry > 0):
            v = 0 if self.__host_len__ == 1 else random.randint(
                0, self.__host_len__-1)
            result = self.__host_cnf__[self.__host__[v]]
            result["serverId"] = self.__host__[v]
            if result['status']:
                break
            retry -= 1
        return result

    def get_host_by_key(self, key):
        return self.__host_cnf__[key]

    def register(self, serverId, host, port: int = 80, sniffer="/api/sniffer", proto: str = "http://", status=True):
        """register
            register an rpc host
        Parameters:
            self - <class: HostManager>, hostManager instance.
            serverId - <class: int>, serverId
            host - <class: str>, ip or domain
            port - <class: int>, port. default 80
            sniffer - <class: str>, the healthy sniffer uri. default /api/sniffer
            proto - <class: str>, http:// or https://. default http://
            status - <class: bool>, default True
        Returns:
            None
        Raises:
            10000 - key already exists.
        """
        serverId = str(serverId)
        proto = self.__checken_proto__(proto)
        if serverId not in self.__host_cnf__:
            self.__host_cnf__[serverId] = {
                "proto": proto,
                "host": host,
                "port": port,
                "status": status,
                "sniffer": sniffer
            }
            if serverId not in self.__host_cnf_backup__:
                self.__host_cnf_backup__ = {
                    "proto": proto,
                    "host": host,
                    "port": port,
                    "status": status,
                    "sniffer": sniffer
                }
            self.__host__.append(serverId)
            self.__host_len__ = len(self.__host__)
        else:
            raise (Exception("10000, '%s' already exists." % serverId))

    def __checken_proto__(self, v):
        v = v.replace(":", "")
        v = v.replace("/", "")
        v = v+"://"
        return v

    def reload(self):
        '''reload
            reload host conf by host conf backup.
        '''
        self.__host_cnf__ = {}
        self.__host__ = []
        self.__host_len__ = 0
        # self.__health_sniffer__ = ""
        for k, v in self.__host_cnf_backup__.items():
            self.register(k, v["host"], v["port"],
                          v["sniffer"], v["proto"], v["status"])

    def delete(self, serverId):
        """delete
            delete a host cnf.
        Parameters:
            self - <class, HostManager>, hostmanager instance.
            serverId - <class, int|str>, serverId.
        Returns:
            success - serverId
            fail - None
        Raises:
            None.
        """
        if str(serverId) in self.__host_cnf__:
            m_result = self.__host_cnf__.pop(serverId)
            self.__host__.remove(serverId)
            self.__host_len__ = len(self.__host__)
            return m_result
        else:
            return None

    def health(self):
        """health
            check host health status.
        Parameters:
            self - <class, HostManager>, hostmanager instance.
        Returns:
            <class, json>, cheked result.
            json desc
                statistics - <class, json>, checked result.
                    total - host count.
                    health - checked it's success count.
                    unhealthy - checked it's fail count.
                health - <class, list>, health serverId list.
                unhealthy - <class, list>, unhealthy serverId list.
        Raises:
            None.
        """
        result = {
            "statistics": {
                "total": self.__host_len__,
                "health": 0,
                "unhealthy": 0,
            },
            "health": [],
            "unhealthy": []
        }
        for k in self.__host__:
            m_uri = self.__generateHost__(self.__host_cnf__[k])
            if m_uri:
                m_uri += "/" + \
                    self.__host_cnf__[k]["sniffer"].strip(
                        '/').replace('\\', "/")
                status, body = GET(m_uri, time_out=5)
                if status == 200:
                    self.__host_cnf__[k]['status'] = True
                    result["health"].append(k)
                else:
                    self.__host_cnf__[k]["status"] = False
                    result["unhealthy"].append(k)
            else:
                self.__host_cnf__[k]["status"] = False
                result['unhealthy'].append(k)
        for k in result["unhealthy"]:
            self.delete(k)
        result["statistics"]["health"] = len(result["health"])
        result["statistics"]["unhealthy"] = len(result["unhealthy"])
        return result

    def __generateHost__(self, uri_cnf: dict):
        proto = uri_cnf["proto"]
        host = uri_cnf["host"]
        port = str(uri_cnf["port"])
        if uri_cnf["status"]:
            if port != 80 or port != "80":
                return proto+host+":"+port
            else:
                return proto+host
        else:
            return None
